// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Implementation specific definitions for Internet Protocol Version 6.
//
// Things we want visible to other kernel modules, yet aren't part of the
// official specifications (i.e. are implementation specific) go here.
// 


#ifndef IP6IMP_INCLUDED
#define IP6IMP_INCLUDED 1

#include <ip6exp.h>

#define IP_NET_STATUS 0
#define IP_HW_STATUS 1

#define PROTOCOL_ANY 0  // BUGBUG: Shouldn't be defined here.


//
// Structure of packet data we pass around the stack.
//
typedef struct IPv6Packet {
    struct IPv6Packet *Next;  // Next entry on a list of packets.
    uint Position;            // Current logical offset into packet.
    void *Data;               // Current pointer into packet data.
    uint ContigSize;          // Amount of contiguous data remaining.
    uint TotalSize;           // Total amount of data remaining.
    IPv6Header *IP;           // IP header for this packet.
    PNDIS_PACKET NdisPacket;  // Original NDIS Packet (if any).
    void *Auxiliary;          // Extra memory allocated by our stack.
    uint Flags;               // Various, see below.
    uint IPPosition;          // Offset at which IP header resides.
} IPv6Packet;

// Flags for above.
#define PACKET_OURS            0x01  // We alloc'd IPv6Packet struct off heap.
#define PACKET_MEDIA_MULTICAST 0x02  // Received as a link-level multicast.
#define PACKET_REASSEMBLED     0x04  // Arrived as a bunch of fragments.
#define PACKET_AUTHENTICATED   0x08  // Passed authentication.
#define PACKET_JUMBO_OPTION    0x10  // Packet has a jumbo payload option.


//
// Macro for comparing IPv6 addresseses.
//
#define IP6_ADDR_EQUAL(x,y) \
        (RtlCompareMemory(x, y, sizeof(IPv6Addr)) == sizeof(IPv6Addr))


//
// The actual definition of a route cache entry
// can be found in route.h.
//
typedef struct RouteCacheEntry RouteCacheEntry;


//
// Structure of a packet context.
//
typedef struct Packet6Context {
    PNDIS_PACKET pc_link;                     // For lists of packets.
    uint pc_offset;                           // Offset where real data starts.
    void (*CompletionHandler)(PNDIS_PACKET);  // Called on event completion.
    void *CompletionData;                     // Data for completion handler.
    IPv6Addr DiscoveryAddress;                // Source address for ND.
} Packet6Context;


//
// The ProtocolReserved field (extra bytes after normal Packet fields)
// is structured as a Packet6Context.
//
// NB: Only packets created by IPv6 have an IPv6 Packet6Context.
// Packets that NDIS hands up to us do NOT have a Packet6Context.
//
#define PC(Packet) ((Packet6Context *)(Packet)->ProtocolReserved)


//
// Global variables exported by the IPv6 driver for use by other
// NT kernel modules.
//
extern NDIS_HANDLE IPv6PacketPool, IPv6BufferPool;


//
// Functions exported by the IPv6 driver for use by other
// NT kernel modules.
//

void
IPv6RegisterULProtocol(uchar Protocol, void *RecvHandler);

extern void
IPv6SendComplete(void *, PNDIS_PACKET, NDIS_STATUS);

extern int
IPv6Receive(void *, IPv6Packet *);

extern void
IPv6ReceiveComplete(void);


#endif // IP6IMP_INCLUDED
