// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// IPv6 fragmentation/reassembly definitions.
//


#ifndef FRAGMENT_H_INCLUDED
#define FRAGMENT_H_INCLUDED 1

//
// IPv6 Fragment Header Format.
// See RFC 1883, page 19.
//
typedef struct FragmentHeader {
    uchar NextHeader;
    uchar Reserved;
    ushort OffsetFlag;  // Offset is upper 13 bits, flag is lowest bit.
    ulong Id;
} FragmentHeader;

#define FRAGMENT_OFFSET_MASK 0xfff8
#define FRAGMENT_FLAG_MASK 0x0001

#define FRAG_HDR_LEN sizeof(FragmentHeader)
#define IPv6_HDR_LEN sizeof(IPv6Header)

//
// Structure used just to link packets together.
// If NDIS Packets had a link field, we wouldn't need this...
//
// REVIEW: Rename this?
//
typedef struct PacketShim {
    struct PacketShim *Next;    // Next packet on list.
    uchar *This;
    ushort Len;
    ushort Offset;
} PacketShim;


//
// Structure used to keep track of the fragments
// being reassembled into a single IPv6 datagram.
//
// REVIEW: Some of these fields are bigger than they need to be.
//
struct Reassembly {
    struct Reassembly *Next;  // Next entry on outstanding partials list.
    IPv6Header IPHdr;         // Copy of original IP header.
    NetTableEntry *RecvNTE;   // NTE that first fragment arrived on.
    ulong Id;                 // Unique (along w/ addrs) datagram identifier.
    uchar *UnfragData;        // Pointer to unfragmentable data.
    ushort UnfragmentLength;  // Length of the unfragmentable part.
    ushort Timer;             // Time to expiration in ticks (see IPv6Timeout).
    uint DataLength;          // Length of the fragmentable part.
    PacketShim *ContigList;   // List of contiguous fragments (from start).
    PacketShim *ContigEnd;    // Last shim on ContigList (for quick access).
    PacketShim *GapList;      // List of non-contiguous fragments.
    ushort Marker;            // The current marker for contiguous data.
    uchar NextHeader;         // Header type following the fragment header.
};

#define DEFAULT_REASSEMBLY_TIMEOUT IPv6TimerTicks(60)  // 60 seconds.


Reassembly *FragmentLookup(IPv6Addr *Source, IPv6Addr *Dest, ulong Id);
void RemoveFromReassemblyList(Reassembly *reass);
void PrintFragHdr(FragmentHeader *frag);
void ReassemblyTimeout();

#endif // FRAGMENT_H_INCLUDED


