// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Exported Neighbor Discovery definitions and declarations.
//


#ifndef NEIGHBOR_INCLUDED
#define NEIGHBOR_INCLUDED

extern ULONGLONG
CalcReachableTime(uint BaseReachableTime);

extern void
NeighborCacheInit(Interface *IF);

extern void
NeighborCacheTimeout(Interface *IF);

extern NeighborCacheEntry *
FindOrCreateNeighbor(Interface *IF, IPv6Addr *Addr);

extern NeighborCacheEntry *
CreatePermanentNeighbor(Interface *IF, IPv6Addr *Addr, void *LinkAddress);

extern void
DADTimeout(NetTableEntry *NTE);

extern void
RouterSolicitSend(Interface *IF);

extern void
RouterSolicitTimeout(Interface *IF);

extern int
RouterSolicitReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code);                // Value of Code field in ICMPv6 header.

extern int
RouterAdvertReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code);                // Value of Code field in ICMPv6 header.

extern int
NeighborSolicitReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code);                // Value of Code field in ICMPv6 header.

extern int
NeighborAdvertReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code);                // Value of Code field in ICMPv6 header.

extern int
RedirectReceive(
    NetTableEntry *RecvNTE,     // NTE that received the packet.
    IPv6Packet *Packet,         // Packet handed to us by ICMPv6Receive.
    uchar Type,                 // Value of Type field in ICMPv6 header.
    uchar Code);                // Value of Code field in ICMPv6 header.

#endif  // NEIGHBOR_INCLUDED
