// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Test tool for sending different types raw IPv6 packets.
//

#ifndef TESTSEND_INCLUDED
#define TESTSEND_INCLUDED 1

#define IP_PROTOCOL_HOP_BY_HOP 0
#define IP_PROTOCOL_ROUTING 43
#define IP_PROTOCOL_FRAGMENT 44
#define IP_PROTOCOL_ICMPv6 58
#define IP_PROTOCOL_NONE   59
#define IP_PROTOCOL_DEST_OPTS 60

#define IP_PROTOCOL_BOGUS  30

//
// Protocol table.
typedef struct ProtoTblStruct {
    char *Name;
    uint ProtoValue;
    int (*cfunc)();
} ProtoTblStruct;


typedef struct MesTypeTblStruct {
    char *Name0;
    char *Name1;
    ushort Value;
    ushort IPProto;
} MesgTypeTblStruct;

typedef struct ErrorTypeTblStruct {
    char *Name0;
    char *Name1;
    ushort Value;
} ErrorTypeTblStruct;


BOOLEAN param(char **strarg, char **argv, int argc, int current);

// Extension header flags.
#define HOP   1
#define DEST  2
#define FRAG  4
#define FRAG0 8
#define RTG   16

// Error flags.
#define PARAM_PROB   0x1
#define TOO_BIG      0x2
#define DEST_UNREACH 0x4
#define TIME_EXCEED  0x8


ProtoTblStruct *GetProtoFunc(char *proto_name);
uint CreateICMPMsg(uchar *Buff, ushort TypeFlag, ushort ErrorType, int *Len);
uint CreateIPMsg(uchar *Buff, ushort TypeFlag, ushort ErrorType, int *Len);
void usage();
void err(char *s);
int parse_addr(char *s, struct sockaddr_in6 *sin6);
BOOLEAN verify_type(char *type, ushort *Flag);
BOOLEAN verify_error(char *type, ushort *Flag);
BOOLEAN param(char **strarg, char **argv, int argc, int current);


//------------ from tcpip6\ipv6\icmp.h -------------------
//
// ICMPv6 Header.
// The actual message body follows this header and is type-specific.
//
typedef struct ICMPv6Header {
    uchar Type;       // Type of message (high bit zero for error messages).
    uchar Code;       // Type-specific differentiater.
    ushort Checksum;  // Calculated over ICMPv6 message and IPv6 psuedo-header.
} ICMPv6Header;

#define IPv6_HDR_LEN 40

//
// ICMPv6 Type field definitions.
//
#define ICMPv6_DESTINATION_UNREACHABLE    1
#define ICMPv6_PACKET_TOO_BIG             2
#define ICMPv6_TIME_EXCEEDED              3
#define ICMPv6_PARAMETER_PROBLEM          4

#define ICMPv6_ECHO_REQUEST               128
#define ICMPv6_ECHO_REPLY                 129
#define ICMPv6_MULTICAST_LISTENER_QUERY   130
#define ICMPv6_MULTICAST_LISTENER_REPORT  131
#define ICMPv6_MULTICAST_LISTENER_DONE    132

#define ICMPv6_ROUTER_SOLICIT             133
#define ICMPv6_ROUTER_ADVERT              134
#define ICMPv6_NEIGHBOR_SOLICIT           135
#define ICMPv6_NEIGHBOR_ADVERT            136
#define ICMPv6_REDIRECT                   137

#define ICMPv6_INFORMATION_TYPE(type)   ((type) & 0x80)
#define ICMPv6_ERROR_TYPE(type)         (((type) & 0x80) == 0)


//
// ICMPv6 Code field definitions.
//

// For Destination Unreachable errors:
#define ICMPv6_NO_ROUTE_TO_DESTINATION          0
#define ICMPv6_COMMUNICATION_PROHIBITED         1
#define ICMPv6_NOT_NEIGHBOR                     2
#define ICMPv6_ADDRESS_UNREACHABLE              3
#define ICMPv6_PORT_UNREACHABLE                 4

// For Time Exceeded errors:
#define ICMPv6_HOP_LIMIT_EXCEEDED               0
#define ICMPv6_REASSEMBLY_TIME_EXCEEDED         1

// For Parameter Problem errors:
#define ICMPv6_ERRONEOUS_HEADER_FIELD           0
#define ICMPv6_UNRECOGNIZED_NEXT_HEADER         1
#define ICMPv6_UNRECOGNIZED_OPTION              2

//
//  Neighbor Discovery message option definitions.
//
#define ND_OPTION_SOURCE_LINK_LAYER_ADDRESS 1
#define ND_OPTION_TARGET_LINK_LAYER_ADDRESS 2
#define ND_OPTION_PREFIX_INFORMATION 3
#define ND_OPTION_REDIRECTED_HEADER 4
#define ND_OPTION_MTU 5

//
//  Neighbor Advertisement message flags.
//
#define ND_NA_FLAG_ROUTER    0x80000000
#define ND_NA_FLAG_SOLICITED 0x40000000
#define ND_NA_FLAG_OVERRIDE  0x20000000

//
// An ICMPv6 error tuple
//
typedef struct ICMPv6MsgType {
    uchar Type;
    uchar Code;
    uint Value;            // holds "unused", "pointer", or "mtu" value
} ICMPv6MsgType;


//------------ from tcpip6\ipv6\fragment.h -------------------
//
// IPv6 Fragment Header Format.
// See RFC 1883, page 19.
//
typedef struct FragmentHeader {
    uchar NextHeader;
    uchar Reserved;
    ushort OffsetFlag;  // Offset is upper 13 bits, flag is lowest bit.
    ulong Id;
} FragmentHeader;

#define FRAGMENT_OFFSET_MASK 0xfff8
#define FRAGMENT_FLAG_MASK 0x0001
#define FRAG_HDR_LEN sizeof(FragmentHeader)


//------------ from tcpip6\inc\ip6opts.h -------------------

//
// Generic Extension Header, with padding.
// 
typedef struct ExtensionHeaderPadded {
    uchar NextHeader;
    uchar HeaderExtLength;  // In 8-byte units, not counting first 8.
    uchar Pad0;            
    uchar Pad1;             // pad header to 4 bytes.
    uint  Pad2;             // pad header to 8 bytes. dangerous.
} ExtensionHeaderPadded;

// 
#if 0
// Generic Extension Header.
typedef struct ExtensionHeader {
    uchar NextHeader;
    uchar HeaderExtLength;  // In 8-byte units, not counting first 8.
} ExtensionHeader;
#endif

//
// Generic Options Header 
// 
typedef struct OptionHeader {
    uchar Type;
    uchar DataLength;  // In 8-byte units, not counting first 8.
} OptionHeader;

//
// Multi-byte pad option header.
//
typedef struct IPv6PadNOption {
    uchar Type;        // Must be 1 for PadN option.
    uchar DataLength;  // Length of the data that follows.
} IPv6PadNOption;

//
// Format of the router alert within the Hop-by-Hop Option header.
//
typedef struct IPv6RouterAlertOption {
    uchar Type;
    uchar Length;
    ushort Value;
} IPv6RouterAlertOption;

typedef struct Options {
    uint JumboLength;  // Length of packet excluding IPv6 header.
    uint Alert;        // REVIEW: for now, buffer pointer to the Alert option
} Options;

// Option Header Values
#define OPT6_PAD_1 0            // Single byte pad
#define OPT6_PAD_N 1            // Multiple byte pad
#define OPT6_JUMBO_PAYLOAD 194  // Jumbo payload (greater than 65,535 bytes)
#define OPT6_ROUTER_ALERT 199   // BUGBUG: this is bogus, waiting on IANA

// Type of actions to be taken with unrecognized header options
#define OPT6_ACTION(a)  ((a) & 0xc0)    // get action bits
#define OPT6_A_SKIP     0x00            // Skip and continue
#define OPT6_A_DISCARD  0x40            // Discard packet
#define OPT6_A_SEND_ICMP_ALL  0x80      // Send ICMP regardless of source addr
#define OPT6_A_SEND_ICMP_NON_MULTI 0xc0 // Send ICMP if non-multicast src addr

// Various constants
#define EXT_HDR_LEN  sizeof(ExtensionHeader) 
#define OPT_HDR_LEN  sizeof(OptionHeader)    
#define EXT_LEN_UNIT 8          // 8-bytes units used for extension hdr length
#define DEFAULT_EXT_HDR_LEN 8   // (bytes) minimum/unaccouted length of 
                                // Destination and Hop-by-Hop headers

#endif  // TESTSEND__INCLUDED
